from django.urls import re_path

from .feeds import GalleryFeed
from .views import (
    GalleryListView,
    GalleryMonthArchiveView,
    GalleryDetailsView,
    TaggedPhotoListView,
    ExifPhotoListView,
    PhotoDetailsView,
)


app_name = "pgallery"

urlpatterns = [
    re_path(route=r"^$", view=GalleryListView.as_view(), name="gallery_list"),
    re_path(route=r"^rss/$", view=GalleryFeed(), name="gallery_rss"),
    re_path(
        route=r"^photo/(?P<pk>\d+)/$",
        view=PhotoDetailsView.as_view(),
        name="photo_details",
    ),
    re_path(
        route=r"^tag/(?P<tag>[ \w]+)/$",
        view=TaggedPhotoListView.as_view(),
        name="tagged_photo_list",
    ),
    re_path(
        route=r"^exif/(?P<exif_key>[ \w]+)/(?P<exif_value>.+)/$",
        view=ExifPhotoListView.as_view(),
        name="exif_photo_list",
    ),
    re_path(
        route=r"^(?P<slug>[-\w]+)/$",
        view=GalleryDetailsView.as_view(),
        name="gallery_details",
    ),
    re_path(
        route=r"^(?P<year>\d{4})/(?P<month>\d+)/$",
        view=GalleryMonthArchiveView.as_view(),
        name="month_archive",
    ),
]



from django.db import models, migrations
import markitup.fields
import model_utils.fields
import django.contrib.postgres.fields
import django.utils.timezone
from django.conf import settings
import django.contrib.postgres.fields.hstore


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Gallery",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        verbose_name="created",
                        editable=False,
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        verbose_name="modified",
                        editable=False,
                    ),
                ),
                (
                    "status",
                    model_utils.fields.StatusField(
                        default="draft",
                        max_length=100,
                        verbose_name="status",
                        no_check_for_status=True,
                        choices=[("draft", "draft"), ("published", "published")],
                    ),
                ),
                (
                    "status_changed",
                    model_utils.fields.MonitorField(
                        default=django.utils.timezone.now,
                        verbose_name="status changed",
                        monitor="status",
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="title")),
                (
                    "slug",
                    models.SlugField(unique=True, max_length=255, verbose_name="slug"),
                ),
                (
                    "description",
                    markitup.fields.MarkupField(
                        no_rendered_field=True, verbose_name="description"
                    ),
                ),
                (
                    "shot_date",
                    models.DateField(null=True, verbose_name="shot date", blank=True),
                ),
                ("_description_rendered", models.TextField(editable=False, blank=True)),
                (
                    "author",
                    models.ForeignKey(
                        editable=False,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="author",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "ordering": ["-shot_date"],
                "verbose_name_plural": "Galleries",
            },
        ),
        migrations.CreateModel(
            name="Photo",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        verbose_name="created",
                        editable=False,
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        verbose_name="modified",
                        editable=False,
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="title")),
                (
                    "image",
                    models.ImageField(
                        upload_to="photos/%Y/%m/%d", verbose_name="image"
                    ),
                ),
                (
                    "tags",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=64), size=None
                    ),
                ),
                (
                    "exif",
                    django.contrib.postgres.fields.hstore.HStoreField(
                        default={}, editable=False, db_index=True
                    ),
                ),
                (
                    "author",
                    models.ForeignKey(
                        editable=False,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="author",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "gallery",
                    models.ForeignKey(
                        related_name="photos",
                        verbose_name="gallery",
                        to="pgallery.Gallery",
                        null=True,
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "ordering": ["created"],
                "verbose_name_plural": "Photos",
            },
        ),
    ]



"""
Administration for photos and galleries.
"""

from django.contrib import admin
from django.db.models import Count
from django.utils.translation import gettext_lazy as _

from .forms import PhotoForm
from .models import Gallery, Photo


class PhotoInline(admin.TabularInline):
    """
    Administration for photos.
    """

    model = Photo
    form = PhotoForm
    ordering = ["created"]

    def get_extra(self, request, obj=None, **kwargs):
        return 0 if obj else 3


class GalleryAdmin(admin.ModelAdmin):
    """
    Administration for galleries.
    """

    list_display = (
        "author",
        "title",
        "status",
        # Having "description" here raises SystemCheckError (admin.E108).
        # We need to remove description from list_display for Django 2.1-2.2
        # See https://code.djangoproject.com/ticket/30543
        # "description",
        "shot_date",
        "modified",
        "photo_count",
    )
    list_display_links = ("title",)
    list_editable = ("status",)
    list_filter = ("status",)
    date_hierarchy = "shot_date"
    prepopulated_fields = {"slug": ("title",)}
    inlines = [PhotoInline]

    def photo_count(self, obj):
        return obj.photo_count

    photo_count.short_description = _("Photo count")

    def get_queryset(self, request):
        """
        Add number of photos to each gallery.
        """
        qs = super().get_queryset(request)
        return qs.annotate(photo_count=Count("photos"))

    def save_model(self, request, obj, form, change):
        """
        Set currently authenticated user as the author of the gallery.
        """
        obj.author = request.user
        obj.save()

    def save_formset(self, request, form, formset, change):
        """
        For each photo set it's author to currently authenticated user.
        """
        instances = formset.save(commit=False)
        for instance in instances:
            if isinstance(instance, Photo):
                instance.author = request.user
            instance.save()


admin.site.register(Gallery, GalleryAdmin)



from django import forms
from django.contrib.postgres.forms import SimpleArrayField
from django.utils.translation import gettext_lazy as _

from .models import Photo


class PhotoForm(forms.ModelForm):
    tags = SimpleArrayField(forms.CharField(), label=_("tags"), required=False)

    class Meta:
        model = Photo
        fields = ["title", "image", "tags"]



from django.conf import settings
from django.contrib.postgres.fields import ArrayField, HStoreField
from django.db import connection, models
from django.db.models.query import QuerySet
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from markitup.fields import MarkupField
from model_utils import Choices
from model_utils.models import StatusModel, TimeStampedModel
from PIL import ExifTags, Image


def sanitize_exif_value(key, value):
    if isinstance(value, str):
        return value.replace("\x00", "").strip()
    return str(value)


class GalleryQuerySet(QuerySet):
    def published(self):
        return self.filter(status="published")


class Gallery(StatusModel, TimeStampedModel):
    STATUS = Choices(("draft", _("draft")), ("published", _("published")))
    author = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        verbose_name=_("author"),
        on_delete=models.CASCADE,
    )
    title = models.CharField(_("title"), max_length=255)
    slug = models.SlugField(_("slug"), max_length=255, unique=True)
    description = MarkupField(_("description"))
    shot_date = models.DateField(_("shot date"), null=True, blank=True)
    cover_photo = models.ImageField(
        _("cover photo"), upload_to="photos/%Y/%m/%d", null=True, blank=True
    )

    objects = GalleryQuerySet.as_manager()

    class Meta:
        verbose_name_plural = _("Galleries")
        ordering = ["-shot_date"]

    def __str__(self):
        """
        The string representation of a gallery is its title.
        """
        return self.title

    def get_absolute_url(self):
        return reverse("pgallery:gallery_details", kwargs={"slug": self.slug})

    def get_teaser_photos(self):
        return self.photos.all()[:4]


class PhotoManager(models.Manager):
    def tagged(self, tag):
        return self.filter(tags__contains=[tag]).order_by("-gallery__shot_date")

    def for_exif(self, exif_key, exif_value):
        return self.filter(exif__contains={exif_key: exif_value}).order_by(
            "-gallery__shot_date"
        )

    def popular_tags(self, count=10):
        query = f"""
        select
            t.tag,
            count(t.tag) as tag_count
        from
            (select unnest(tags) as tag from {Photo._meta.db_table}) t
        group by tag
        order by tag_count desc
        limit %s
        """
        cursor = connection.cursor()
        cursor.execute(query, (count,))
        tags = [{"tag": row[0], "count": row[1]} for row in cursor.fetchall()]
        return tags


class Photo(TimeStampedModel):
    gallery = models.ForeignKey(
        Gallery,
        null=True,
        related_name="photos",
        verbose_name=_("gallery"),
        on_delete=models.CASCADE,
    )
    author = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        verbose_name=_("author"),
        on_delete=models.CASCADE,
    )
    title = models.CharField(_("title"), max_length=255)
    image = models.ImageField(_("image"), upload_to="photos/%Y/%m/%d")
    tags = ArrayField(models.CharField(max_length=64), blank=True, default=list)
    exif = HStoreField(editable=False, default=dict, db_index=True)

    objects = PhotoManager()

    class Meta:
        verbose_name_plural = _("Photos")
        ordering = ["created"]

    def __str__(self):
        """
        The string representation of a photo is its title.
        """
        return self.title

    def save(self, *args, **kwargs):
        """
        Updates EXIF data before saving.
        """
        # you really should be doing this in a background task
        try:
            img = Image.open(self.image.file)
            raw_exif = img._getexif()
            if raw_exif:
                self.exif = {
                    ExifTags.TAGS[k]: sanitize_exif_value(k, v)
                    for k, v in raw_exif.items()
                    if k in ExifTags.TAGS
                }
        except Exception:
            pass
        super().save(*args, **kwargs)

    def get_absolute_url(self):
        return reverse("pgallery:photo_details", kwargs={"pk": self.pk})

    def get_next_photo(self):
        """
        Returns next photo from the same gallery (in chronological order).

        Wraps around from last photo in the gallery to the first one.
        """
        try:
            next_photo = Photo.objects.filter(
                gallery=self.gallery, created__gt=self.created
            )[0]
        except IndexError:
            next_photo = Photo.objects.filter(gallery=self.gallery)[0]
        return next_photo

    def get_previous_photo(self):
        """
        Returns previous photo from the same gallery (in chronological order).

        Wraps around from first photo in the gallery to the last one.
        """
        try:
            previous_photo = Photo.objects.filter(
                gallery=self.gallery, created__lt=self.created
            ).latest("created")
        except Photo.DoesNotExist:
            previous_photo = Photo.objects.filter(gallery=self.gallery).latest(
                "created"
            )
        return previous_photo



from django.contrib.sitemaps import Sitemap

from .models import Gallery, Photo


class GallerySitemap(Sitemap):
    changefreq = "weekly"
    priority = 0.5

    def items(self):
        return Gallery.published.all()

    def lastmod(self, obj):
        return obj.modified


class PhotoSitemap(Sitemap):
    changefreq = "weekly"
    priority = 0.5

    def items(self):
        return Photo.objects.all()

    def lastmod(self, obj):
        return obj.modified



# Generated by Django 1.9.5 on 2016-04-16 23:51

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pgallery", "0003_gallery_cover_photo"),
    ]

    operations = [
        migrations.AlterField(
            model_name="photo",
            name="tags",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=64),
                blank=True,
                default=list,
                size=None,
            ),
        ),
    ]



from django import template

from pgallery.models import Gallery, Photo

register = template.Library()


@register.simple_tag
def get_recent_galleries(count=3):
    """
    Returns most recent galleries.
    """
    return Gallery.objects.published().order_by("-shot_date")[:count]


@register.simple_tag
def get_gallery_archive_dates():
    """
    Returns datetime objects for all months in which galleries were added.
    """
    return Gallery.objects.published().dates("shot_date", "month", order="DESC")


@register.simple_tag
def get_popular_tags(count=10):
    """
    Returns most popular tags.
    """
    return Photo.objects.popular_tags(count)



"""
Gallery syndication feeds.
"""

from django.contrib.syndication.views import Feed
from django.urls import reverse_lazy
from django.utils.translation import gettext as _

from .models import Gallery


class GalleryFeed(Feed):
    """
    RSS feed with latest galleries.
    """

    title = _("siciarz.net - galleries")
    link = reverse_lazy("pgallery:gallery_list")
    description = _("Latest photo galleries")

    def items(self):
        return Gallery.objects.published()[:5]

    def item_title(self, item):
        return item.title

    def item_description(self, item):
        return item.description



# Generated by Django 2.2.12 on 2020-04-12 11:11

import django.contrib.postgres.fields.hstore
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("pgallery", "0004_auto_20160416_2351"),
    ]

    operations = [
        migrations.AlterField(
            model_name="photo",
            name="exif",
            field=django.contrib.postgres.fields.hstore.HStoreField(
                db_index=True, default=dict, editable=False
            ),
        ),
    ]



# Generated by Django 1.9.2 on 2016-02-07 12:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pgallery", "0002_auto_20150417_1442"),
    ]

    operations = [
        migrations.AddField(
            model_name="gallery",
            name="cover_photo",
            field=models.ImageField(
                blank=True,
                null=True,
                upload_to="photos/%Y/%m/%d",
                verbose_name="cover photo",
            ),
        ),
    ]



from django.db import models, migrations
import django.contrib.postgres.fields


class Migration(migrations.Migration):

    dependencies = [
        ("pgallery", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="photo",
            name="tags",
            field=django.contrib.postgres.fields.ArrayField(
                default=[],
                size=None,
                base_field=models.CharField(max_length=64),
                blank=True,
            ),
        ),
    ]



from django.views.generic import DetailView, ListView, MonthArchiveView

from .models import Gallery, Photo


class StaffAccessMixin:
    """
    Allow staff members to see all galleries, including drafts.
    """

    def get_queryset(self):
        if self.request.user.is_staff:
            return Gallery.objects.all()
        return Gallery.objects.published()


class GalleryListView(StaffAccessMixin, ListView):
    pass


class GalleryMonthArchiveView(StaffAccessMixin, MonthArchiveView):
    date_field = "shot_date"
    month_format = "%m"
    make_object_list = True


class GalleryDetailsView(StaffAccessMixin, DetailView):
    pass


class TaggedPhotoListView(ListView):
    def get_queryset(self):
        return Photo.objects.tagged(self.kwargs["tag"])

    def get_context_data(self, **kwargs):
        data = super().get_context_data(**kwargs)
        data["tag"] = self.kwargs["tag"]
        return data


class ExifPhotoListView(ListView):
    template_name = "pgallery/exif_photo_list.html"

    def get_queryset(self):
        return Photo.objects.for_exif(
            self.kwargs["exif_key"], self.kwargs["exif_value"]
        )

    def get_context_data(self, **kwargs):
        data = super().get_context_data(**kwargs)
        data["exif_key"] = self.kwargs["exif_key"]
        data["exif_value"] = self.kwargs["exif_value"]
        return data


class PhotoDetailsView(DetailView):
    model = Photo
